// CONJGRAD.CPP

// Copyright (C) 1998 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

////////////////////////////////////////////////////////////////////////////////////////////////////

#include "conjgrad.h"

////////////////////////////////////////////////////////////////////////////////////////////////////

conjugate_gradient::conjugate_gradient(ls_mode p1, f32 p2, i32s p3)
{
	mode = p1; defstp = p2;
	reset = p3; step = 0;
	
	delta = 0.001;	// how to set optimal value??? this is just the default...
}

conjugate_gradient::~conjugate_gradient(void)
{
}

void conjugate_gradient::AddVar(f32 * p1, f32 * p2)
{
	cgvar newvar; newvar.ref1 = p1; newvar.ref2 = p2;
	cgvar_vector.push_back(newvar);
}

// Haataja J, Kpyaho J, Rahola J : "Numeeriset menetelmt", CSC-Tieteellinen laskenta OY, 1993

bool conjugate_gradient::TakeCGStep(void)
{
	optstp = 0.0;
	optval = GetGradient();
	
	// compute inner product of the latest gradient to newip.
	
	newip = 0.0;
	for (i32u n1 = 0;n1 < cgvar_vector.size();n1++)
	{
		f32 tmp1 = (* cgvar_vector[n1].ref2);
		newip += tmp1 * tmp1;
	}
	
	// either reset beta or update it, using the Fletcher-Reeves method.
	
	if (!(step++ % reset))
	{
		beta = 0.0;
		for (i32u n1 = 0;n1 < cgvar_vector.size();n1++)
		{
			cgvar_vector[n1].data1 = 0.0;
		}
	}
	else
	{
		if (oldip == 0.0) return false;
		beta = newip / oldip;
	}
	
	oldip = newip;
	
	// store the current search direction and the current value for a while...
	
	f32 sum = 0.0;
	for (i32u n1 = 0;n1 < cgvar_vector.size();n1++)
	{
		cgvar_vector[n1].data2 = (* cgvar_vector[n1].ref1);
		f32 tmp1 = cgvar_vector[n1].data1 * beta - (* cgvar_vector[n1].ref2);
		cgvar_vector[n1].data1 = tmp1; sum += tmp1 * tmp1;
	}
	
	f32 scale = sqrt(sum);
	f32 stpln = defstp / scale;
	
	// do the line search...
	
	if (mode == Newton2An)		// f'/f''-newton half numerical
	{
		i32s newton = 0;
		while (true)
		{
			InitLineSearch(stpln);
			f32 tmp1 = GetGradient();
			
			if (tmp1 < optval)
			{
				optstp = stpln;
				optval = tmp1;
			}
			
			if (newton++ > 3) break;
			
			f32 delta = stpln * 0.001, tmp2 = 0.0, tmp3 = 0.0;
			for (i32u n1 = 0;n1 < cgvar_vector.size();n1++)
			{
				tmp2 += cgvar_vector[n1].data1 * (* cgvar_vector[n1].ref2) / scale;
			}
			
			InitLineSearch(stpln + delta);
			GetGradient();
			
			for (i32u n1 = 0;n1 < cgvar_vector.size();n1++)
			{
				tmp3 += cgvar_vector[n1].data1 * (* cgvar_vector[n1].ref2) / scale;
			}
			
			f32 denom = tmp3 - tmp2;
			if (denom != 0.0) stpln = fabs(stpln - delta * tmp2 / denom);
			else break;
		}
	}
	
	if (mode == Newton2Num)		// f'/f''-newton full numerical
	{
		i32s newton = 0;
		while (true)
		{
			InitLineSearch(stpln);
			f32 tmp1 = GetValue();
			
			if (tmp1 < optval)
			{
				optstp = stpln;
				optval = tmp1;
			}
			
			if (newton++ > 3) break;
			
			f32 delta = stpln * 0.001;
			
			InitLineSearch(stpln + delta);
			f32 tmp2 = GetValue();
			
			InitLineSearch(stpln + delta * 2.0);
			f32 tmp3 = GetValue();
			
			f32 denom = tmp3 - 2.0 * tmp2 + tmp1;
			if (denom != 0.0) stpln = fabs(stpln - delta * (tmp2 - tmp1) / denom);
			else break;
		}
	}
	
	if (mode == Newton1Num)		// f/f'-newton full numerical
	{
		i32s newton = 0;
		while (true)
		{
			InitLineSearch(stpln);
			f32 tmp1 = GetValue();
			
			if (fabs(tmp1) < fabs(optval))
			{
				optstp = stpln;
				optval = tmp1;
			}
			
			if (newton++ > 3) break;
			
			f32 delta = stpln * 0.001;
			
			InitLineSearch(stpln + delta);
			f32 tmp2 = GetValue();
			
			f32 denom = tmp2 - tmp1;
			if (denom != 0.0) stpln = fabs(stpln - delta * tmp1 / denom);
			else break;
		}
	}
	
	if (mode == SimpleTest)		// a simple test method
	{
		InitLineSearch(stpln);
		if (GetValue() < optval) optstp = stpln * 1.05;
		else optstp = stpln * 0.1;
	}
	
	// finally take the optimal step and scale back.
	
	InitLineSearch(optstp);
	optstp *= scale;
	return true;
}

void conjugate_gradient::InitLineSearch(f32 p1)
{
	for (i32u n1 = 0;n1 < cgvar_vector.size();n1++)
	{
		f32 tmp1 = cgvar_vector[n1].data2 + cgvar_vector[n1].data1 * p1;
		(* cgvar_vector[n1].ref1) = tmp1;
	}
}

// this will calculate numerical derivatives (in forward direction). can be replaced by
// analytical derivative code if needed. must always return the same value as GetValue()!!!!!

// how to set optimal value for delta?????????????

f32 conjugate_gradient::GetGradient(void)
{
	f32 value = GetValue();
	for (i32u n1 = 0;n1 < cgvar_vector.size();n1++)
	{
		f32 old = (* cgvar_vector[n1].ref1);
		(* cgvar_vector[n1].ref1) = old + delta;
		(* cgvar_vector[n1].ref2) = (GetValue() - value) / delta;
		(* cgvar_vector[n1].ref1) = old;
	}
	
	return value;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

// eof
