// MODEL.H : a set of base classes common to all models.

// Copyright (C) 1998 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

////////////////////////////////////////////////////////////////////////////////////////////////////

// if we study the design of this software and wish to name some central feature that would
// give the best brief description of this design, I think that feature should be "the separation
// of data storage and computation". by "data storage" I mean the description of the system under
// study: in MM models that means the data about atoms and bonds, and in QM models that might mean
// data about atoms and the state of the system (total charge, spin state???). this data is common
// to all models that belong to the same category (MM, QM) and should be separated from the
// individual implementations of different computational techniques (the forcefield X, the quantum-
// chemical theory Y). this is because the graphical user interfaces are dependent on those
// descriptions of the systems, and not on those computational details.

// practically this division is made in the following way: we call those data storage classes as
// "model"-classes, and those classes that implement the computational details as "engine"-classes.
// the "model" class always decribes the system under study. when we want to compute something, we
// create an instance of some suitable "engine"-class using our "model"-class. the "engine"-class
// will copy that information it needs, and calculates those results it is supposed to calculate.
// if we calculate some useful results that change our original system, we can copy those results
// back to our "model"-class.

// and here's the beef: nothing prevents us from doing as many separate "engine"-classes as we
// need or wish for our N+1 forcefields and theories!!!

////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef MODEL_H
#define MODEL_H

class bs_mdl_simple;
class bs_mdl_extended;

class crd_set;
class superimpose;

////////////////////////////////////////////////////////////////////////////////////////////////////

#include "conjgrad.h"

#include BEGIN_INCLUDE_STL	// this is compiler-dependent...
#include <vector>
#include END_INCLUDE_STL	// this is compiler-dependent...

////////////////////////////////////////////////////////////////////////////////////////////////////

// "simple" models can handle only a single conformation...
// "simple" models can handle only a single conformation...
// "simple" models can handle only a single conformation...

class bs_mdl_simple
{
	public:
	
	// general-purpose output stream and buffer.
	// if "ostr" is NULL no output is allowed/needed...
	
	ostream * ostr; char buffer[1024];
	
	public:
	
	bs_mdl_simple(ostream *);
	virtual ~bs_mdl_simple(void);
};

////////////////////////////////////////////////////////////////////////////////////////////////////

// "extended" models can handle multiple conformations...
// "extended" models can handle multiple conformations...
// "extended" models can handle multiple conformations...

class bs_mdl_extended : public bs_mdl_simple
{
	public:
	
	vector<crd_set> cs_vector;
	
	public:
	
	bs_mdl_extended(ostream *);
	virtual ~bs_mdl_extended(void);
	
	void UpdateAccumValues(void);
	
	// it is supposed that at least one coordinate set exists all the time!!!
	// it is supposed that at least one coordinate set exists all the time!!!
	// it is supposed that at least one coordinate set exists all the time!!!
	
	virtual void PushCRDSets(i32u) = 0;
	virtual void PopCRDSets(i32u) = 0;
	
	virtual void CopyCRDSet(i32u, i32u) = 0;
	virtual void SwapCRDSets(i32u, i32u) = 0;
	
	virtual void CenterCRDSet(i32u) = 0;
	virtual void ReserveCRDSets(i32u) = 0;
};

////////////////////////////////////////////////////////////////////////////////////////////////////

// on some models we want to be able to handle several conformations (or "coordinate sets")
// simultaneously. currently "crd-set" class handles this...

// store topology and coordinates in separate files??? maybe in MM1 but what about others???
// store topology and coordinates in separate files??? maybe in MM1 but what about others???
// store topology and coordinates in separate files??? maybe in MM1 but what about others???

class crd_set
{
	public:
	
	char * description;
	
// bring also the coloring information here??? -> would allow different colors for different crdsets!!!
// bring also the coloring information here??? -> would allow different colors for different crdsets!!!
// bring also the coloring information here??? -> would allow different colors for different crdsets!!!
	
	f32 accum_weight;
	fGL accum_value;
	bool visible;
	
	public:
	
	crd_set(void);
	crd_set(const crd_set &);
	~crd_set(void);
	
	void SetDescription(const char *);
};

////////////////////////////////////////////////////////////////////////////////////////////////////

class superimpose : public conjugate_gradient
{
	protected:
	
	i32s index[2];
	
	f32 rot[3]; f32 drot[3];
	f32 loc[3]; f32 dloc[3];
	
	public:
	
	superimpose(i32s, i32s);
	~superimpose(void);
	
	void GetMatrix(fGL *) const;
	
	virtual void Transform(void) const = 0;
};

////////////////////////////////////////////////////////////////////////////////////////////////////

#endif	// MODEL_H

// eof
