// NEURAL.CPP

// Copyright (C) 1999 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

////////////////////////////////////////////////////////////////////////////////////////////////////

#include "neural.h"

#include <math.h>

#include <time.h>
#include <stdlib.h>

#include <iostream.h>
#include <fstream.h>

////////////////////////////////////////////////////////////////////////////////////////////////////

bp_network::bp_network(i32s isz, i32s hsz, i32s osz)
{
	size[0] = isz; size[1] = hsz; size[2] = osz;
	
	iw = new f32 *[size[0]];
	for (i32s n1 = 0;n1 < size[0];n1++)
	{
		iw[n1] = new f32[size[1]];
		for (i32s n2 = 0;n2 < size[1];n2++) iw[n1][n2] = 0.0;
	}
	
	hw = new f32 *[size[1]];
	for (i32s n1 = 0;n1 < size[1];n1++)
	{
		hw[n1] = new f32[size[2]];
		for (i32s n2 = 0;n2 < size[2];n2++) hw[n1][n2] = 0.0;
	}
	
	ha = new f32[size[1]]; oa = new f32[size[2]];
	hd = new f32[size[1]]; hs = new f32[size[1]];
}

bp_network::~bp_network(void)
{
	for (i32s n1 = 0;n1 < size[0];n1++) delete[] iw[n1]; delete[] iw;
	for (i32s n1 = 0;n1 < size[1];n1++) delete[] hw[n1]; delete[] hw;
	delete[] ha; delete[] oa; delete[] hd; delete[] hs;
}

void bp_network::Randomize(f32 range)
{
	srand(time(NULL));
	
	for (i32s n1 = 0;n1 < size[0];n1++)
	{
		for (i32s n2 = 0;n2 < size[1];n2++)
		{
			f32 rndnum = range * ((f32) rand() / (f32) RAND_MAX - 0.5);
			iw[n1][n2] = rndnum;
		}
	}
	
	for (i32s n1 = 0;n1 < size[1];n1++)
	{
		for (i32s n2 = 0;n2 < size[2];n2++)
		{
			f32 rndnum = range * ((f32) rand() / (f32) RAND_MAX - 0.5);
			hw[n1][n2] = rndnum;
		}
	}
}

void bp_network::ReadNetwork(const char * fn)
{
	ifstream file(fn, ios::in);
	
	i32s sz; bool error = false;
	file >> sz; if (sz != size[0]) error = true;
	file >> sz; if (sz != size[1]) error = true;
	file >> sz; if (sz != size[2]) error = true;
	
	if (error)
	{
		cout << "incompatible file -> failed!!!" << endl;
		file.close(); return;
	}
	
	char buffer[256]; f32 value;  
	file.getline(buffer, sizeof(buffer));
	
	for (i32s n1 = 0;n1 < size[0];n1++)
	{
		for (i32s n2 = 0;n2 < size[1];n2++)
		{
			file >> value;
			iw[n1][n2] = value;
		}
		
		file.getline(buffer, sizeof(buffer));
	}
	
	for (i32s n1 = 0;n1 < size[1];n1++)
	{
		for (i32s n2 = 0;n2 < size[2];n2++)
		{
			file >> value;
			hw[n1][n2] = value;
		}
		
		file.getline(buffer, sizeof(buffer));
	}
	
	file.close();
}

void bp_network::WriteNetwork(const char * fn)
{
	ofstream file(fn, ios::out);
	
	file << size[0] << "   ";
	file << size[1] << "   ";
	file << size[2] << endl;
	
	for (i32s n1 = 0;n1 < size[0];n1++)
	{
		for (i32s n2 = 0;n2 < size[1];n2++)
		{
			file << iw[n1][n2] << " ";
		}
		
		file << endl;
	}
	
	for (i32s n1 = 0;n1 < size[1];n1++)
	{
		for (i32s n2 = 0;n2 < size[2];n2++)
		{
			file << hw[n1][n2] << " ";
		}
		
		file << endl;
	}
	
	file.close();
}

void bp_network::Predict(const f32 * id, f32 * od)
{
	for (i32s n1 = 0;n1 < size[1];n1++)
	{
		ha[n1] = 0.0;
		for (i32s n2 = 0;n2 < size[0];n2++)
		{
			ha[n1] += iw[n2][n1] * id[n2];
		}
		
		hd[n1] = Sigmoid(ha[n1]);
	}
	
	for (i32s n1 = 0;n1 < size[2];n1++)
	{
		oa[n1] = 0.0;
		for (i32s n2 = 0;n2 < size[1];n2++)
		{
			oa[n1] += hw[n2][n1] * hd[n2];
		}
		
		od[n1] = Sigmoid(oa[n1]);
	}
}

void bp_network::Learn(const f32 * id, f32 * od, const f32 * target, f32 alpha)
{
	Predict(id, od);
	for (i32s n1 = 0;n1 < size[1];n1++) hs[n1] = 0.0;
	
	for (i32s n1 = 0;n1 < size[2];n1++)
	{
		f32 delta = DSigmoid(oa[n1]) * (target[n1] - od[n1]);
		
		for (i32s n2 = 0;n2 < size[1];n2++)
		{
			hs[n2] += delta * hw[n2][n1];
			hw[n2][n1] += alpha * delta * hd[n2];
		}
	}
	
	for (i32s n1 = 0;n1 < size[1];n1++)
	{
		f32 delta = DSigmoid(ha[n1]) * hs[n1];
		
		for (i32s n2 = 0;n2 < size[0];n2++)
		{
			iw[n2][n1] += alpha * delta * id[n2];
		}
	}
}

f32 bp_network::Sigmoid(f32 a)
{
	f32 tmp1 = exp(-a);
	f32 tmp2 = 1.0 + tmp1;
	
	return (1.0 / tmp2);
}

f32 bp_network::DSigmoid(f32 a)
{
	f32 tmp1 = exp(-a);
	f32 tmp2 = 1.0 + tmp1;
	
	return (tmp1 / (tmp2 * tmp2));
}

////////////////////////////////////////////////////////////////////////////////////////////////////

// eof
