// VIEWING.H : some derived viewing/utility classes.

// Copyright (C) 1998 Tommi Hassinen, Jarno Huuskonen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef VIEWING_H
#define VIEWING_H

#ifndef TARGET1

class camera;

class light;
class spot_light;
class directional_light;

//class clipper; ???
//class single_clipper;
//class double_clipper;

class transformer;

////////////////////////////////////////////////////////////////////////////////////////////////////

class window;	// window.h
class bs_gui;	// guibase.h

#include "objects.h"

////////////////////////////////////////////////////////////////////////////////////////////////////

// "camera" is a class which holds information about position and orientation of the views
// in viewing windows. a single camera can pass this information to multiple windows. there
// can also be "local lights", lights connected to a camera. if some translations/rotations
// are done to the camera, these lights will follow...

class camera : public dummy_object
{
	public:
	
	fGL focus;
	bs_gui * gui;
	
	bool use_local_lights;
	bool use_global_lights;
	bool ortho;
	
	public:
	
	enum rmode { Normal = 0, Draw = 1, Select = 2 };
	
	camera(const object_location &, fGL, bs_gui *);
	camera(const camera &);
	~camera(void);
	
	void RenderWindow(window *, rmode = Normal, int = 0, int = 0);
	void RenderObjects(window *);
	
	const char * GetObjectName(void) { return NULL; }	// virtual
	
	void OrbitObject(const fGL *, const camera &);		// virtual
	void RotateObject(const fGL *, const camera &);		// virtual
	
	void TranslateObject(const fGL *, const obj_loc_data *);	// virtual
	
	bool BeginTransformation(void) { return false; }	// virtual
	bool EndTransformation(void) { return false; }		// virtual
	
	void Render(void);		// virtual
};

////////////////////////////////////////////////////////////////////////////////////////////////////

struct light_components
{
	fGL * amb_comp;
	fGL * diff_comp;
	fGL * spec_comp;
};

// "light" is a base class for OpenGL lights. all lights must be dummy_objects, because they are
// connected to a camera rather than to a set of windows. that's because cameras will drag their
// local lights along when they are rotated/translated...

// also there is no need for lights to respond to any connectivity/geometry changes...

class light : public light_components, public dummy_object
{
	public:
	
// "owner" should be either NULL (-> global light) or some valid camera of the model (-> local light).
// only model::Add???Light()-functions are allowed to make modifications to "owner" !!!

	camera * owner;
	
// "number" is the OpenGL light number GL_LIGHTi for this light.
// the actual values are determined in the model when adding/removing local/global lights...

	iGLs number;
	
	static const fGL def_amb_comp[4];
	static const fGL def_diff_comp[4];
	static const fGL def_spec_comp[4];
	
	static const light_components def_components;
	
	public:
	
	light(const object_location &, const light_components &);
	virtual ~light(void);
	
// "SetupProperties" is used to set up those light properties that are independent on
// location/orientation -> should be called for every window when it's created and always when
// the light numbering is changed.

// "SetupLocation" should be called for all active lights always when rendering them,
// after the modelview-matrix is ready but before any actual rendering commands.

	virtual void SetupProperties(void) = 0;
	virtual void SetupLocation(void) = 0;
	
	bool BeginTransformation(void) { return false; }	// virtual
	bool EndTransformation(void) { return false; }		// virtual
	
	private:
	
	void InitComponents(const light_components *);
};

////////////////////////////////////////////////////////////////////////////////////////////////////

// the OpenGL spotlight (a light source at finite distance).
// it has the standard cutoff/exponent-properties, and will be rendered as a sphere/cone...

class spot_light : public light
{
	public:
	
	fGL cutoff;
	fGL exponent;
	
	static fGL size;
	
	static fGL shade1[3];
	static fGL shade2[3];
	
	static fGL bulb_on[3];
	static fGL bulb_off[3];
	
	static const fGL def_cutoff;
	static const fGL def_exponent;
	
	public:
	
	spot_light(const object_location &, const light_components & = light::def_components, fGL = spot_light::def_cutoff, fGL = spot_light::def_exponent);
	virtual ~spot_light(void);
	
	const char * GetObjectName(void) { return "Light (spotlight)"; }	// virtual
	
	void SetupProperties(void);		// virtual
	void SetupLocation(void);		// virtual
	
	void Render(void);		// virtual
};

////////////////////////////////////////////////////////////////////////////////////////////////////

// the OpenGL directional light (a light source at infinite distance). no rendering for this one...

class directional_light : public light
{
	protected:
	
	public:
	
	directional_light(const object_location &, const light_components & = light::def_components);
	virtual ~directional_light(void);
	
	const char * GetObjectName(void) { return "Light (directional)"; }	// virtual
	
	void SetupProperties(void);		// virtual
	void SetupLocation(void);		// virtual
	
	void Render(void);		// virtual
};

////////////////////////////////////////////////////////////////////////////////////////////////////

// "transformer" is a special object used only when rotating/translating some selected part of
// the model. transformer::GetMatrix() is used to get the transformation matrix for making those
// modifications permanent.

// it is useful to switch orbiting and rotating, so that objects will behave similarly to lights
// and the whole scene...

class transformer : public dummy_object
{
	public:
	
	bs_gui * target_model;
	
	static bool transform_in_progress;
	
	friend class mm1_gui;	// temporary !?!?!?!?
	
	public:
	
	transformer(void);
	~transformer(void);
	
	void Init(bs_gui *);
	void GetMatrix(fGL *) const;
	
	const char * GetObjectName(void) { return NULL; }	// virtual
	
	bool BeginTransformation(void);		// virtual
	bool EndTransformation(void);		// virtual
	
	void OrbitObject(const fGL *, const camera &);		// virtual
	void RotateObject(const fGL *, const camera &);		// virtual
	
	void Render(void);	// virtual
};

////////////////////////////////////////////////////////////////////////////////////////////////////

#endif	// TARGET1
#endif	// VIEWING_H

// eof
